module picoampmeter (
  input CLK, //16Mhz Clock
  input DDC_DOUT, //Serial output of DDC
  input DDC_NDVA, //Valid data, active low

  output DDC_TEST, //test DDC with charge injection
  output DDC_CONV, //wich side of integrator is converting
  output DDC_CLK, //clk of DDC112 (10MHz nominal)
  output DDC_DCLK, //data clock for transmission
  output DDC_NDX, //enable serial, active low
  output DDC_RA0, //range lsb
  output DDC_RA1, //range
  output DDC_RA2, //range msb

  output USBPU,  // USB pull-up resistor
  output PIN_1 // Serial TX
);
// drive USB pull-up resistor to '0' to disable USB
assign USBPU = 0;

//Uart module
reg enableUartTx; //Enable uart transmitter
reg [7:0] UartByte; //Byte to send trough uart
wire UartDone;
parameter c_CLKS_PER_BIT = 1667; //Baudrate 9600, 16'000'000/9600 = 1667
uart_tx #(.CLKS_PER_BIT(c_CLKS_PER_BIT)) UART_TX_INST
(.i_Clock(CLK),
 .i_Tx_DV(enableUartTx),
 .i_Tx_Byte(UartByte),
 .o_Tx_Active(),
 .o_Tx_Serial(PIN_1),
 .o_Tx_Done(UartDone)
 );

//Select range of integrator
// 001 => 0 to 50pC
assign DDC_RA0 = 1;
assign DDC_RA1 = 1;
assign DDC_RA2 = 1;

//Generate 8MHz clock for DDC CLK
reg clk_8;
always @(posedge CLK) begin
  clk_8 <= !clk_8; //Divide clk by 2
end
assign DDC_CLK = clk_8;

parameter integration_time_us = 15999999;
//Generate integration signal (conv)
reg conv;
reg [23:0] counter = integration_time_us; //1s integration
always @ (posedge CLK) begin
  if (counter == 0)
    begin
      conv <= !conv;
      counter <= integration_time_us;
    end
  else
    counter <= counter - 1;
end
assign DDC_CONV = conv;

//Generate test signal
assign DDC_TEST = 1'b0; //continuous test (13 pC per integration)

//Read serial data
parameter s_IDLE                 = 3'b000;
parameter s_READDDC              = 3'b001;
parameter s_CHECK_READDC_IS_DONE = 3'b010;
parameter s_TRANSFERUART         = 3'b011;
parameter s_WAITUART             = 3'b100;
parameter s_DONE                 = 3'b101;

reg [2:0] StateMachine = 0;
reg dclk;
reg [39:0] buffer;
reg [5:0] bufferpt;
reg r_ndx;
always @ (posedge CLK) begin
  case(StateMachine)
    s_IDLE :
      begin
        r_ndx <= 1'b1; //disable ndx
        dclk <= 1'b0; //disable dclk
        bufferpt <= 39; //start with msb (big-endian)
        if (DDC_NDVA == 1'b0) //Data valid, read serial data output
          begin
            r_ndx <= 1'b0; //Take ndx low to acknoledge data valid
            StateMachine <= s_READDDC;
          end
        else
          StateMachine <= s_IDLE;
      end

    s_READDDC :
      begin

        if (dclk == 1'b1)
          begin
            buffer[bufferpt] <= DDC_DOUT;
            bufferpt <= bufferpt - 1;
          end
        dclk <= !dclk; //toggle clock

        StateMachine <= s_CHECK_READDC_IS_DONE;
      end

    s_CHECK_READDC_IS_DONE :
      begin

        if (bufferpt < 40)
          StateMachine <= s_READDDC;
        else
          begin
            bufferpt <= 0;
            UartByte <= 10; //newline
            enableUartTx <= 1'b1;
            StateMachine <= s_WAITUART;
          end
        // Display Allo! on UART (debug)
        /*bufferpt <= 0;
        buffer[7:0] <= 65;
        buffer[15:8] <= 108;
        buffer[23:16] <= 108;
        buffer[31:24] <= 111;
        buffer[39:32] <= 33;

        // Display 13pC on both channel
        bufferpt <= 0;
        buffer[39:0] <= 40'b0100001010001111010101000010100011110101;
        UartByte <= 10; //newline
        enableUartTx <= 1'b1;
        StateMachine <= s_WAITUART;
        */
      end

    s_WAITUART :
      begin
        if (UartDone == 1'b1)
          StateMachine <= s_TRANSFERUART;
        else
          begin
            enableUartTx <= 1'b0;
            StateMachine <= s_WAITUART;
          end
      end

    s_TRANSFERUART :
      begin
        if (bufferpt < 39)
          begin
            UartByte <= {buffer[bufferpt+7],
                          buffer[bufferpt+6],
                          buffer[bufferpt+5],
                          buffer[bufferpt+4],
                          buffer[bufferpt+3],
                          buffer[bufferpt+2],
                          buffer[bufferpt+1],
                          buffer[bufferpt]};
            bufferpt <= bufferpt + 8;
            enableUartTx <= 1'b1;
            StateMachine <= s_WAITUART;
          end
        else
          begin
            StateMachine <= s_IDLE;
          end
      end

    s_DONE :
      begin
        StateMachine <= s_DONE;
      end

  endcase
  end
  assign DDC_DCLK = dclk;
  assign DDC_NDX = r_ndx;
endmodule // picoampmeter
